# Coordinate ID generator:
# Takes a set of 3d coordinates (x, y, z), truncates to 2dp and then
# generates a base64-like(*) representation that can be used as an id
# for the object.
#
# The actual encoding is not base64, since base64 contains characters
# that present a problem when used, e.g, in URIs. I replaced the '+'
# and '/' characters with '_' and '.' respectively.
#
# Example usage:
#   coord_to_id64(51.5625, 32.1875, -27.125)
# Generates:
#   "1gA:Oi:-Go"
#
# A reversing function is also provided.
#
# Original Author: Oliver "kfsone" Smith <oliver@kfs.org>
# Released under the "use it with attribution" license.

from __future__ import print_function, division
import string

alphabet = string.digits + string.ascii_lowercase + string.ascii_uppercase + '_.'
precision = 100.

def coord_to_d64(coord):
    i = int(abs(coord * precision))

    digits = ""
    while True:
        digits = alphabet[i & 63] + digits
        i >>= 6   # divides by 64, or one digit
        if i == 0:
            break

    sign = '-' if coord < 0 else ''
    return sign + digits


def d64_to_coord(d64):
    divisor, digits = precision, d64
    if d64.startswith('-'):
        divisor = -divisor
        digits = digits[1:]

    number = 0
    for digit in digits:
        value = alphabet.find(digit)
        if value < 0:
            raise ValueError("Invalid d64 value: {}".format(value))
        number = (number * 64) + value

    return number / divisor


def pos_to_id64(x, y, z):
    return coord_to_d64(x) + ':' + coord_to_d64(y) + ':' + coord_to_d64(z)


def id64_to_pos(id64):
    (x64, y64, z64) = id64.split(':')
    return (d64_to_coord(x64), d64_to_coord(y64), d64_to_coord(z64))


if __name__ == "__main__":
    test1 = ( 51.5625,32.1875,-27.125 )
    test2 = ( -154.65625,40.34375,-82.78125 )

    id64 = pos_to_id64(test1[0], test1[1], test1[2])
    print("id64 of {} = {}".format(test1, id64))
    pos = id64_to_pos(id64)
    print("pos of {} = {}".format(id64, pos))

    id64 = pos_to_id64(test1[0], test1[1], test1[2])
    print("id64 of {} = {}".format(test1, id64))
    pos = id64_to_pos(id64)
    print("pos of {} = {}".format(id64, pos))

    id64 = pos_to_id64(test2[0], test2[1], test2[2])
    print("id64 of {} = {}".format(test2, id64))
    pos = id64_to_pos(id64)
    print("pos of {} = {}".format(id64, pos))



from .. import cache
from .. import corrections
from .. import csvexport
import os
import pathlib
import platform
from .. import plugins
import re
import time
from .. import tradedb
from .. import tradeenv
from .. import transfers

from ..plugins import PluginException

# Constants

BASE_URL = "http://www.davek.com.au/td/"
CORRECTIONS_URL = BASE_URL + "correctionsfile.asp"
SYSTEMS_URL = BASE_URL + "System.csv"
STATIONS_URL = BASE_URL + "station.asp"
SHIPVENDOR_URL = BASE_URL + "shipvendor.asp"

class DecodingError(PluginException):
    pass


class ImportPlugin(plugins.ImportPluginBase):
    """
    Plugin that downloads data from maddavo's site.
    """

    stampFile = "maddavo.stamp"
    dateRe = re.compile(r"(\d\d\d\d-\d\d-\d\d)[ T](\d\d:\d\d:\d\d)")

    pluginOptions = {
        'csvs':         "Merge Corrections, System, Station and ShipVendor "
                        "data into the local db.",
        'corrections':  "Merge Corrections data into local db.",
        'systems':      "Merge System data into local db.",
        'stations':     "Merge Station data into local db.",
        'shipvendors':  "Merge ShipVendor data into local db.",
        'exportcsv':    "Regenerate System and Station .csv files after "
                        "merging System/Station data.",
        'csvonly':      "Stop after csv work, don't import prices",
        'skipdl':       "Skip doing any downloads.",
        'force':        "Process prices even if timestamps suggest "
                        "there is no new data.",
        'use3h':        "Force download of the 3-hours .prices file.",
        'use2d':        "Force download of the 2-days .prices file.",
        'usefull':      "Force download of the full .prices file.",
    }


    def __init__(self, tdb, tdenv):
        super().__init__(tdb, tdenv)

        self.filename = self.defaultImportFile
        stampFilePath = pathlib.Path(ImportPlugin.stampFile)
        self.stampPath = tdb.dataPath / stampFilePath
        self.modSystems = 0
        self.modStations = 0


    def load_timestamp(self):
        """
        Read the date from the timestamp file.
        Returns a zero date if the file doesn't exist or
        doesn't contain a date.
        """

        prevImportDate = None
        lastRunDays = float("inf")
        if self.stampPath.is_file():
            try:
                fh = self.stampPath.open('rU', encoding="utf-8")
                line = fh.readline().split('\n')
                if line and line[0]:
                    if ImportPlugin.dateRe.match(line[0]):
                        prevImportDate = line[0]
                line = fh.readline().split('\n')
                if line and line[0]:
                    lastRunAge = time.time() - float(line[0])
                    lastRunDays = lastRunAge / (24 * 60 * 60)
            except FileNotFoundError:
                pass

        if not prevImportDate:
            prevImportDate = "0000-00-00 00:00:00"
        return prevImportDate, lastRunDays


    def save_timestamp(self, newestDate, startTime):
        """
        Save a date to the timestamp file.
        """

        with self.stampPath.open('w', encoding="utf-8") as fh:
            print(newestDate, file=fh)
            print(startTime, file=fh)


    def check_shebang(self, line, checkAge):
        m = re.match(
            r'^#!\s*trade.py\s*import\s*.*\s*--timestamp\s*"([^"]+)"',
            line
        )
        if not m:
            raise PluginException(
                "Data is not Maddavo's prices list format: " + line
            )
        self.importDate = m.group(1)
        if checkAge and not self.getOption("force"):
            if self.importDate <= self.prevImportDate:
                raise SystemExit(
                    "Local data is already current [{}]."
                    .format(self.importDate)
                )
            if self.tdenv.detail:
                print("New timestamp: {}, Old timestamp: {}".format(
                    self.importDate,
                    self.prevImportDate
                ))


    def csv_stream_rows(self, url, tableName):
        """
        Iterate rows from a CSV resource for a given table.
        """
        if self.getOption(tableName.lower()):
            self.tdenv.NOTE("Importing {}", tableName)
            for _, values in transfers.CSVStream(url, tdenv=self.tdenv):
                yield values


    def import_corrections(self):
        """
        Fetch and import deletions/renames from Corrections.csv
        """

        tdb, tdenv = self.tdb, self.tdenv
        sysLookup = tdb.systemByName.get
        db = tdb.getDB()
        deletes, updates = 0, 0

        # Build a name lookup table from a given ID index.
        def make_index(table):
            return { i.dbname: i for i in table.values() }

        # For each ID table that we accept corrections for, build
        # a name-based lookup index.
        basicTypes = {
            'Category': ('category_id', make_index(tdb.categoryByID)),
            'Item': ('item_id', make_index(tdb.itemByID)),
            'Rare': ('rare_id', make_index(tdb.rareItemByID)),
            'Ship': ('ship_id', make_index(tdb.shipByID)),
            'System': ('system_id', make_index(tdb.systemByID)),
            'Station': ('station_id', {
                stn.name().upper(): stn for stn in tdb.stationByID.values()
            }),
        }

        tables = {
            'System':   corrections.systems,
            'Station':  corrections.stations,
            'Category': corrections.categories,
            'Item':     corrections.items,
        }

        FIXING, DELETING, DISCARDING = 'FIXING', 'DELETING', 'DISCARDING'

        stream = self.csv_stream_rows(CORRECTIONS_URL, "Corrections")
        for src, oldName, newName in stream:
            action = FIXING if (newName != 'DELETED') else DELETING

            try:
                (idColumn, index) = basicTypes[src]
            except KeyError:
                tdenv.NOTE("Unsupported correction type {} ignored", src)
                continue

            correctionTable = tables.get(src, None)
            if correctionTable:
                if action is DELETING:
                    correction = corrections.DELETED
                else:
                    correction = newName
                correctionTable[oldName.upper()] = correction

            item = index.get(oldName, None)
            if not item:
                tdenv.DEBUG1(
                    "Correction for {} not needed, skipping.", oldName
                )
                continue

            if src == 'Station':
                if action is FIXING:
                    newItemName = "/".join(
                        [item.system.dbname, newName]
                    ).upper()
            else:
                if action is FIXING:
                    newItemName = newName

            if action is FIXING:
                newItem = index.get(newItemName, None)
                if newItem and newItem is not item:
                    tdenv.DEBUG1(
                        "{} exists so {} can be deleted.",
                        newItemName, oldName,
                    )
                    action = DISCARDING
                    if correctionTable:
                        correctionTable[oldName.upper()] = corrections.DELETED

            if action is FIXING:
                tdenv.DEBUG0("{} {} {} -> {}", action, src, oldName, newName)
                stmt = "UPDATE {} SET name = ? WHERE {} = ?".format(
                    src, idColumn
                )
                binds = [newName, item.ID]
                updates += 1
            else:
                tdenv.DEBUG0("{} {} {}", action, src, oldName)
                stmt = "DELETE FROM {} WHERE {} = ?".format(src, idColumn)
                binds = [item.ID]
                deletes += 1

            if tdenv.debug:
                tdenv.DEBUG1("{} [{}]", stmt, binds)
            db.execute(stmt, binds)

        if updates == 0 and deletes == 0:
            tdenv.DEBUG0("No corrections applied.")
            return

        tdenv.NOTE("{} updates, {} deletions", updates, deletes)
        db.commit()
        db.execute("VACUUM")
        if not self.getOption("exportcsv"):
            tdenv.WARN(
                "Corrections imported without --opt=exportcsv. "
                "You may want to manually export data later on."
            )

        # We need to reload stuff.
        tdb.load(maxSystemLinkLy=tdenv.maxSystemLinkLy)


    def csv_system_rows(self, url, tableName):
        """
        Iterate systems from a CSV resource.
        """
        tdb, tdenv = self.tdb, self.tdenv
        sysLookup = tdb.systemByName.get
        sysAdjust = corrections.systems.get
        DELETED = corrections.DELETED

        stream = self.csv_stream_rows(url, tableName)
        for values in stream:
            srcName = values[0].upper()
            sysName = sysAdjust(srcName, srcName)
            if sysName == DELETED:
                tdenv.DEBUG0("Ignoring deleted system {}", srcName)
                continue
            system = sysLookup(sysName, None)
            yield sysName, system, values


    def import_systems(self):
        """
        Fetch and import data from Systems.csv
        """

        tdb, tdenv = self.tdb, self.tdenv
        dateTimeRe = re.compile(r'\s*(\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2})')

        stream = self.csv_system_rows(SYSTEMS_URL, "Systems")
        for sysName, system, values in stream:
            added = values[4]
            modified = values[5]
            if added.startswith("DEL") or modified.startswith("DEL"):
                if system:
                    tdb.removeLocalSystem(
                        system
                    )
                self.modSystems += 1
                continue

            # Clean up the modified stamp
            m = dateTimeRe.match(modified)
            if not m:
                tdenv.WARN("{} has invalid 'modified': {}", sysName, modified)
                continue
            modified = m.group(1)

            x, y, z = float(values[1]), float(values[2]), float(values[3])
            if not system:
                tdb.addLocalSystem(
                    sysName, x, y, z, added, modified,
                    commit=False,
                )
                self.modSystems += 1
            elif system.posX != x or system.posY != y or system.posZ != z:
                print("{} position change: {}v{}, {}v{}, {}v{}".format(
                    sysName, system.posX, x, system.posY, y, system.posZ, z
                ))
                tdb.updateLocalSystem(
                    system, sysName, x, y, z, added, modified,
                    commit=False,
                )
                self.modSystems += 1

        if self.modSystems:
            tdb.getDB().commit()


    def csv_station_rows(self, url, tableName):
        """
        Fetch rows from a CSV resource that start with a system and a station,
        applying corrections.
        """

        tdb, tdenv = self.tdb, self.tdenv
        sysAdjust = corrections.systems.get
        stnAdjust = corrections.stations.get
        DELETED = corrections.DELETED

        stream = self.csv_system_rows(url, tableName)
        for sysName, system, values in stream:
            if not system:
                tdenv.NOTE(
                    "Unrecognized system for station {}/{}",
                    sysName, stnName,
                )
                continue
            stnName = values[1]
            stnName = stnAdjust('/'.join([sysName, stnName.upper()]), stnName)
            if stnName == DELETED:
                tdenv.DEBUG0(
                    "Ignoring deleted station {}/{}",
                    values[0], values[1],
                )
                continue
            station = system.getStation(stnName)
            yield system, stnName, station, values


    def import_stations(self):
        """
        Fetch and import data from Stations.csv.
        """

        tdb, tdenv = self.tdb, self.tdenv

        stream = self.csv_station_rows(STATIONS_URL, "Stations")
        for system, stnName, station, values in stream:
            try:
                lsFromStar = float(values[2])
            except ValueError:
                tdenv.WARN(
                    "Invalid, non-numeric, lsFromStar value ('{}') for {}/{}",
                    values[2], system.name(), stnName
                )
                continue
            if int(lsFromStar) != lsFromStar:
                tdenv.NOTE(
                    "Discarding floating-point part of {} for {}/{}",
                    lsFromStar, system.name(), stnName
                )
                lsFromStar = int(lsFromStar)
            modified = values[7] if len(values) > 7 else 'now'
            if lsFromStar < 0 or modified.startswith("DEL"):
                if station:
                    tdb.removeLocalStation(station, commit=False)
                    self.modStations += 1
                continue
            blackMarket = values[3]
            maxPadSize = values[4]
            market = values[5] if len(values) > 5 else '?'
            shipyard = values[6] if len(values) > 6 else '?'
            outfitting = values[8] if len(values) > 8 else '?'
            rearm = values[9] if len(values) > 9 else '?'
            refuel = values[10] if len(values) > 10 else '?'
            repair = values[11] if len(values) > 11 else '?'
            planetary = values[12] if len(values) > 12 else '?'
            if station:
                if tdb.updateLocalStation(
                        station,
                        name=stnName,
                        lsFromStar=lsFromStar,
                        market=market,
                        blackMarket=blackMarket,
                        shipyard=shipyard,
                        maxPadSize=maxPadSize,
                        outfitting=outfitting,
                        rearm=rearm,
                        refuel=refuel,
                        repair=repair,
                        planetary=planetary,
                        modified=modified,
                        commit=False,
                        ):
                    self.modStations += 1
            else:
                tdb.addLocalStation(
                    system=system,
                    name=stnName,
                    lsFromStar=lsFromStar,
                    market=market,
                    blackMarket=blackMarket,
                    shipyard=shipyard,
                    maxPadSize=maxPadSize,
                    outfitting=outfitting,
                    rearm=rearm,
                    refuel=refuel,
                    repair=repair,
                    planetary=planetary,
                    modified=modified,
                    commit=False,
                )
                self.modStations += 1

        if self.modStations:
            tdenv.DEBUG0("commit")
            tdb.getDB().commit()


    def import_shipvendors(self):
        """
        Fetch dave's ShipVendor.csv and import it.
        """

        tdb, tdenv = self.tdb, self.tdenv
        db = tdb.getDB()
        ships = { ship.dbname.upper(): ship for ship in tdb.shipByID.values() }
        lastFail = None

        # Index ship locations.
        stmt = """
            SELECT  ship_id, station_id
              FROM  ShipVendor
        """
        shipLocs = set(
            (shipID, stnID) for shipID, stnID in db.execute(stmt)
        )
        newShips = {}
        delShips = set()

        stream = self.csv_station_rows(SHIPVENDOR_URL, "ShipVendors")
        for system, stnName, station, values in stream:
            if not station:
                if not tdenv.quiet:
                    name = "{}/{}".format(system.name(), stnName)
                    if name != lastFail:
                        tdenv.NOTE(
                            "Ignoring unrecognized station: {}/{}",
                            system.name(), stnName,
                        )
                        lastFail = name
                continue
            shipName = values[2]
            modified = values[3] if len(values) > 3 else 'now'
            isDelete = modified.startswith("DEL")
            ship = ships.get(shipName.upper(), None)
            if not ship:
                if not isDelete:
                    tdenv.NOTE(
                        "Ignoring unrecognized ship {} at {}",
                        shipName, station.name()
                    )
                continue
            locKey = (ship.ID, station.ID)
            if locKey in shipLocs:
                if isDelete and not locKey in delShips:
                    try:
                        newShips.remove(locKey)
                    except KeyError:
                        pass
                    delShips.add(locKey)
                    tdenv.NOTE(
                        "Removing {} from {}",
                        ship.name(), station.name()
                    )
                continue
            if locKey in newShips:
                continue
            try:
                delShips.remove(locKey)
            except KeyError:
                pass
            tdenv.NOTE(
                "At {} adding {}", station.name(), ship.name(),
            )
            newShips[locKey] = (ship.ID, station.ID, modified)
        if newShips:
            stmt = """
                REPLACE INTO ShipVendor
                (ship_id, station_id, modified)
                VALUES
                (?, ?, DATETIME(?))
            """
            db.executemany(stmt, newShips.values())
        if delShips:
            db.executemany("""
                DELETE FROM ShipVendor WHERE ship_id = ? and station_id = ?
            """, delShips)
        if newShips or delShips:
            db.commit()


    def refresh_csv(self):
        if not self.getOption("exportcsv"):
            return

        for table in [
            "Category", "Item",
            "System", "Station",
            "Ship", "ShipVendor",
            "RareItem"
        ]:
            _, path = csvexport.exportTableToFile(
                self.tdb, self.tdenv, table
            )
            self.tdenv.NOTE("{} re-exported.", path)

    def download_prices(self, lastRunDays):
        """ Figure out which file to download and fetch it. """

        # Argument checking
        use3h = 1 if self.getOption("use3h") else 0
        use2d = 1 if self.getOption("use2d") else 0
        usefull = 1 if self.getOption("usefull") else 0
        if use3h + use2d + usefull > 1:
            raise PluginException(
                "Only one of use3h/use2d/usefull can be used at once."
            )
        if self.getOption("skipdl"):
            if (use3h or use2d or usefull):
                raise PluginException(
                    "use3h/use2d/usefull has no effect with --opt=skipdl"
                )
            return

        # Overrides
        if use3h:
            lastRunDays = 0.01
        elif use2d:
            lastRunDays = 1.0
        elif usefull:
            lastRunDays = 3.0

        # Use age/options to determine which file
        if lastRunDays < 3 / 24:
            priceFile = "prices-3h.asp"
        elif lastRunDays < 1.9:
            priceFile = "prices-2d.asp"
        else:
            priceFile = "prices.asp"

        # Fetch!
        transfers.download(
            self.tdenv,
            BASE_URL + priceFile,
            self.filename,
            shebang=lambda line: self.check_shebang(line, True),
        )

    def import_prices(self):
        """ Download and import data price data """

        tdb, tdenv = self.tdb, self.tdenv

        # It takes a while to download these files, so we want
        # to record the start time before we download. What we
        # care about is when we downloaded relative to when the
        # files were previously generated.
        startTime = time.time()

        prevImportDate, lastRunDays = self.load_timestamp()
        self.prevImportDate = prevImportDate

        self.download_prices(lastRunDays)
        if tdenv.download:
            return

        # Scan the file for the latest data.
        firstDate = None
        newestDate = prevImportDate
        numNewLines = 0
        minLen = len(prevImportDate) + 10
        dateRe = ImportPlugin.dateRe
        lastStn = None
        updatedStations = set()
        tdenv.DEBUG0("Reading prices data")
        with open(self.filename, "rUb") as fh:
            # skip the shebang.
            firstLine = fh.readline().decode(encoding="utf-8")
            self.check_shebang(firstLine, False)
            importDate = self.importDate

            lineNo = 0
            while True:
                lineNo += 1
                try:
                    line = next(fh)
                except StopIteration:
                    break
                try:
                    line = line.decode(encoding="utf-8")
                except UnicodeDecodeError as e:
                    try:
                        line = line.decode(encoding="latin1")
                        line = line.encode("utf-8")
                        line = line.decode()
                    except UnicodeDecodeError:
                        raise DecodingError(
                            "{} line {}: "
                            "Invalid (unrecognized, non-utf8) character "
                            "sequence: {}\n{}".format(
                                self.filename, lineNo, str(e), line,
                            )
                        ) from None
                    raise DecodingError(
                        "{} line {}: "
                        "Invalid (latin1, non-utf8) character "
                        "sequence:\n{}".format(
                            self.filename, lineNo, line,
                        )
                    )
                if line.startswith('@'):
                    lastStn = line[2:line.find('#')].strip()
                    continue
                if not line.startswith(' ') or len(line) < minLen:
                    continue
                m = dateRe.search(line)
                if not m:
                    continue
                date = m.group(1) + ' ' + m.group(2)
                if not firstDate or date < firstDate:
                    firstDate = date
                if date > prevImportDate:
                    updatedStations.add(lastStn)
                    numNewLines += 1
                    if date > newestDate:
                        newestDate = date
                        if date > importDate:
                            raise PluginException(
                                "Station {} has suspicious date: {} "
                                "(newer than the import?)"
                                .format(lastStn, date)
                            )

        if numNewLines == 0:
            tdenv.NOTE("No new price entries found.")

        forceParse = self.getOption("force") or self.getOption("skipdl")
        if numNewLines > 0 or forceParse:
            if tdenv.detail:
                print(
                    "Date of last import   : {}\n"
                    "Timestamp of import   : {}\n"
                    "Oldest update in file : {}\n"
                    "Newest update in file : {}\n"
                    "Number of new entries : {}\n"
                    .format(
                        prevImportDate,
                        importDate,
                        firstDate,
                        newestDate,
                        numNewLines,
                    ))

            numStationsUpdated = len(updatedStations)
            if not tdenv.quiet and numStationsUpdated:
                if len(updatedStations) > 12 and tdenv.detail < 2:
                    updatedStations = list(updatedStations)[:10] + ["..."]
                tdenv.NOTE(
                    "{} {} updated:\n{}",
                    numStationsUpdated,
                    "stations" if numStationsUpdated > 1 else "station",
                    ', '.join(updatedStations)
                )

            cache.importDataFromFile(
                tdb,
                tdenv,
                pathlib.Path(self.filename),
            )

        self.save_timestamp(importDate, startTime)

    def run(self):
        tdb, tdenv = self.tdb, self.tdenv

        tdenv.ignoreUnknown = True

        if self.getOption("csvs"):
            self.options["corrections"] = True
            self.options["systems"] = True
            self.options["stations"] = True
            self.options["shipvendors"] = True
            self.options["exportcsv"] = True

        # Ensure the cache is built and reloaded.
        tdb.reloadCache()
        tdb.load(maxSystemLinkLy=tdenv.maxSystemLinkLy)

        self.import_corrections()
        self.import_systems()
        self.import_stations()
        self.import_shipvendors()

        # Let the system decide if it needs to reload-cache
        tdb.close()

        self.refresh_csv()

        if not self.getOption("csvonly"):
            self.import_prices()

        tdenv.NOTE("Import completed.")

        # We did all the work
        return False



#! /usr/bin/env python

# Experimental module to generate a JSON version of the .prices file.

# Set to True to allow export of systems that don't have any station data
emptySystems = True
# Set to True to allow export of stations that don't have prices
emptyStations = True

import sqlite3
import json
import time
import collections
import os

conn = sqlite3.connect("data/TradeDangerous.db")

def collectItemData(db):
    """
    Builds a flat, array of item names that serves as a table of items.
    Station Price Data refers to this table by position in the array.
    As a result, we also need a mapping for itemID -> tableID
    """

    items = []      # table of items
    itemIdx = {}    # mapping from itemID -> position in items
    for ID, name in db.execute("SELECT i.item_id, i.name FROM Item AS i"):
        itemIdx[ID] = len(items)
        items.append(name)

    return items, itemIdx


def collectSystems(
            db,
            itemIdx,
            withEmptySystems=False,
            withStations=True,
            withEmptyStations=False,
            withPrices=True
    ):
    """
    Build the System -> Station -> Price data from the supplied DB.

    withEmptySystems:
        True: include systems with no station data.

    withStations:
        False: (implies withEmptySystems=True) don't include station data.

    withEmptyStations:
        True: include stations with no price data.

    withPrices:
        False: (implies withEmptyStations=True) don't include price data.
    """

    systems = collections.defaultdict(dict)
    if not withStations:
        withEmptySystems = True

    for sysID, sys, posX, posY, posZ in db.execute("""
            SELECT  sys.system_id, sys.name,
                    sys.pos_x, sys.pos_y, sys.pos_z
            FROM  System AS sys
                    LEFT OUTER JOIN Station
                        USING (system_id)
            GROUP  BY 1
            """):
        systemData = {
                 'pos': [ posX, posY, posZ ],
        }
        if withStations:
            stations = collectStations(
                    db, itemIdx,
                    sysID,
                    withEmptyStations=withEmptyStations,
                    withPrices=withPrices,
            )
            if not stations and not withEmptySystems:
                continue
            if stations:
                systemData['stn'] = stations
        systems[sys] = systemData

    return systems


def collectStations(
            db,
            itemIdx,
            sysID,
            withEmptyStations=False,
            withPrices=True
    ):
    """
    Populate a station list for a given system, including price data.
    """
    stations = {}
    if not withPrices:
        withEmptyStations = True
    for stnID, name, lsFromStar, lastMod in db.execute("""
            SELECT  stn.station_id, stn.name, stn.ls_from_star,
                    MAX(si.modified)
              FROM  Station AS stn
                    LEFT OUTER JOIN StationItem si
                        USING (station_id)
             WHERE  stn.system_id = ?
             GROUP  BY 1
            """, [sysID]):
        if not lastMod and not withEmptyStations:
            continue
        stationData = {
                'ls':   int(lsFromStar),
        }
        if lastMod:
            stationData['ts'] = lastMod
            if withPrices:
                buy, sell = collectPriceData(
                        db, itemIdx,
                        stnID,
                )
                stationData['buy'] = buy
                stationData['sell'] = sell

        stations[name] = stationData

    return stations


def collectPriceData(db, itemIdx, stnID):
    """
    Collect buying and selling data for a given station. Items reference the
    position in the itemData array.
    """

    buying, selling = [], []

    for itmID, cr in db.execute("""
            SELECT  sb.item_id, sb.price
              FROM  StationBuying AS sb
             WHERE  station_id = ?
            """, [stnID]):
        buying.append([ itemIdx[itmID], cr ])

    for itmID, cr, units, level in db.execute("""
            SELECT  ss.item_id, ss.price, ss.units, ss.level
              FROM  StationSelling AS ss
             WHERE  station_id = ?
            """, [stnID]):
        selling.append([ itemIdx[itmID], cr, units, level ])

    return buying, selling


itemData, itemIdx = collectItemData(conn)
sysData = collectSystems(
        conn, itemIdx,
        withEmptySystems=emptySystems, withEmptyStations=emptyStations
)
jsonData = {
    'src': 'td/json-exp',
    'time': int(time.time()),
    'items': itemData,
    'emptySys': emptySystems,
    'emptyStn': emptyStations,
    'sys': sysData,
}

try:
    cmdrName = os.environ['CMDR']
    jsonData['cmdr'] = cmdrName
except KeyError:
    pass

print(json.dumps(jsonData, separators=(',',':')))




from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from os import getcwd, path
from collections import deque
from pathlib import Path
from .tradeexcept import TradeException

import csv
import json
import math
from .misc import progress as pbar
from . import fs
import platform
import time
import subprocess
import sys

try:
    import requests
    __requests = requests
except ImportError:
    __requests = None

def import_requests():
    global __requests
    global platform
    if __requests:
        return __requests

    if platform.system() == 'Linux':
        extra = (
            "Ubuntu users: You may be able to install 'pip'\n"
            "with 'apt-get install python3-pip' and requests with\n"
            "'pip3 install --upgrade requests'.\n"
        )
    elif platform.system() == 'Windows':
        extra = (
            "\n\nThe requests package can be installed with\n"
            "'pip3 install --upgrade requests'.\n\n"
            "If requests is installed, you may have bits\n"
            "of 32-bit and 64-bit Python installed.\n"
            "Consider using control panel to uninstall Python,\n"
            "delete the Python folder (usually C:\\Python34\\),\n"
            "and then re-install Python.\n"
        )
    else:
        extra = ""

    print(
        "ERROR: Unable to load the Python 'requests' package.\n" + extra
    )
    approval = input(
        "I can try and install the package automatically using 'pip'.\n"
        "Try to install 'requests' now (y/n)? "
    )
    # Idiot-proofing: take just the first character in case the user typed
    # 'YES' (upper or lower case) instead of 'Y'.
    if approval[0:1].lower() != 'y':
        raise TradeException("Missing package: 'requests'")

    try:
        import pip
    except ImportError as e:
        import platform
        raise TradeException(
            "Python 3.4.2 includes a package manager called 'pip', "
            "except it doesn't appear to be installed on your system:\n"
            "{}{}".format(str(e), extra)
        ) from None

    # Let's use "The most reliable approach, and the one that is fully supported."
    # Especially since the old way produces an error for me on Python 3.6:
    # "AttributeError: 'module' object has no attribute 'main'"
    #pip.main(["install", "--upgrade", "requests"])
    subprocess.check_call([sys.executable, '-m', 'pip', 'install', '--upgrade', 'requests'])

    try:
        import requests
        __requests = requests
    except ImportError as e:
        raise TradeException(
            "The requests module did not install correctly.{}"
            .format(extra)
        ) from None

    return __requests


######################################################################
# Helpers

class HTTP404(TradeException):
    pass


def makeUnit(value):
    """
    Convert a value in bytes into a Kb, Mb, Gb etc.
    """

    units = [ 'B ', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB' ]
    unitSize = int(value)
    for unit in units:
        if unitSize <= 640:
            return "{:>5.01f}{}".format(unitSize, unit)
        unitSize /= 1024
    return None


def download(
            tdenv, url, localFile,
            headers=None,
            backup=False,
            shebang=None,
            chunkSize=4096,
        ):
    """
    Fetch data from a URL and save the output
    to a local file. Returns the response headers.

    tdenv:
        TradeEnv we're working under

    url:
        URL we're fetching (http, https or ftp)

    localFile:
        Name of the local file to open.

    headers:
        dict() of additional HTTP headers to send

    shebang:
        function to call on the first line
    """

    requests = import_requests()
    tdenv.NOTE("Requesting {}".format(url))
    req = requests.get(url, headers=headers or None, stream=True)
    req.raise_for_status()

    encoding = req.headers.get('content-encoding', 'uncompress')
    length = req.headers.get('content-length', None)
    transfer = req.headers.get('transfer-encoding', None)
    if transfer != 'chunked':
        # chunked transfer-encoding doesn't need a content-length
        if length is None:
            print(req.headers)
            raise Exception("Remote server replied with invalid content-length.")
        length = int(length)
        if length <= 0:
            raise TradeException(
                "Remote server gave an empty response. Please try again later."
            )

    if tdenv.detail > 1:
        if length:
            tdenv.NOTE("Downloading {} {}ed data", makeUnit(length), encoding)
        else:
            tdenv.NOTE("Downloading {} {}ed data", transfer, encoding)
    tdenv.DEBUG0(str(req.headers).replace("{", "{{").replace("}", "}}"))

    # Figure out how much data we have
    if length and not tdenv.quiet:
        progBar = pbar.Progress(length, 20)
    else:
        progBar = None

    actPath = Path(localFile)
    fs.ensurefolder(tdenv.tmpDir)
    tmpPath = Path(tdenv.tmpDir, "{}.dl".format(actPath.name))

    histogram = deque()

    fetched = 0
    lastTime = started = time.time()
    spinner, spinners = 0, [
        '.    ', '..   ', '...  ', ' ... ', '  ...', '   ..', '    .'
    ]
    with tmpPath.open("wb") as fh:
        for data in req.iter_content(chunk_size=chunkSize):
            fh.write(data)
            fetched += len(data)
            if shebang:
                bangLine = data.decode().partition("\n")[0]
                tdenv.DEBUG0("Checking shebang of {}", bangLine)
                shebang(bangLine)
                shebang = None
            if progBar:
                now = time.time()
                deltaT = max(now - lastTime, 0.001)
                lastTime = now
                if len(histogram) >= 15:
                    histogram.popleft()
                histogram.append(len(data) / deltaT)
                progBar.increment(
                    len(data),
                    postfix=lambda value, goal: \
                            " {:>7s} [{:>7s}/s] {:>3.0f}% {:1s}".format(
                            makeUnit(value),
                            makeUnit(sum(histogram) / len(histogram)),
                            (fetched * 100. / length),
                            spinners[spinner]
                        )
                )
                if deltaT > 0.200:
                    spinner = (spinner + 1) % len(spinners)
        tdenv.DEBUG0("End of data")
    if not tdenv.quiet:
        if progBar:
            progBar.clear()
        elapsed = (time.time() - started) or 1
        tdenv.NOTE(
            "Downloaded {} of {}ed data {}/s",
            makeUnit(fetched), encoding,
            makeUnit(fetched / elapsed)
        )

    fs.ensurefolder(actPath.parent)

    # Swap the file into place
    if backup:
        bakPath = Path(localFile + ".bak")
        if bakPath.exists():
            bakPath.unlink()
        if actPath.exists():
            actPath.rename(localFile + ".bak")
    if actPath.exists():
        actPath.unlink()
    tmpPath.rename(actPath)

    req.close()
    return req.headers


def get_json_data(url):
    """
    Fetch JSON data from a URL and return the resulting dictionary.

    Displays a progress bar as it downloads.
    """

    requests = import_requests()
    req = requests.get(url, stream=True)

    totalLength = req.headers.get('content-length')
    if totalLength is None:
        compression = req.headers.get('content-encoding')
        compression = (compression + "'ed") if compression else "uncompressed"
        print("Downloading {}: {}...".format(compression, url))
        jsData = req.content
    else:
        totalLength = int(totalLength)
        progBar = pbar.Progress(totalLength, 25)
        jsData = bytes()
        for data in req.iter_content():
            jsData += data
            progBar.increment(
                len(data),
                postfix=lambda value, goal: \
                " {}/{}".format(
                    makeUnit(value),
                    makeUnit(goal),
            ))
        progBar.clear()

    return json.loads(jsData.decode())


class CSVStream(object):
    """
    Provides an iterator that fetches CSV data from a given URL
    and presents it as an iterable of (columns, values).

    Example:
        stream = transfers.CSVStream("http://blah.com/foo.csv")
        for cols, vals in stream:
            print("{} = {}".format(cols[0], vals[0]))
    """

    def __init__(self, url, tdenv=None):
        self.url = url
        self.tdenv = tdenv
        if not url.startswith("file:///"):
            requests = import_requests()
            self.req = requests.get(self.url, stream=True)
            self.lines = self.req.iter_lines()
        else:
            self.lines = open(url[8:], "rUb")
        self.columns = self.next_line().split(',')

    def next_line(self):
        """ Fetch the next line as a text string """
        while True:
            line = next(self.lines)
            try:
                return line.decode(encoding="utf-8")
            except UnicodeDecodeError as e:
                if not self.tdenv:
                    raise e
                self.tdenv.WARN(
                    "{}: line:{}: {}\n{}",
                    self.url, self.csvin.line_num, line, e
                )

    def __iter__(self):
        """
        Iterate across data received as csv values.
        Yields [column headings], [column values]
        """
        self.csvin = csvin = csv.reader(
            iter(self.next_line, 'END'),
            delimiter=',', quotechar="'", doublequote=True
        )
        columns = self.columns
        for values in csvin:
            if values and len(values) == len(columns):
                yield columns, values



"""This module should handle filesystem related operations
"""
from shutil import copy as shcopy
from os import makedirs, path, utime
from pathlib import Path

__all__ = ['copy', 'copyallfiles', 'touch', 'ensurefolder']

def pathify(*args):
    if len(args) > 1 or not isinstance(args[0], Path):
        return Path(*args)
    return args[0]


def copy(src, dst):
    """
    copy src to dst and takes string or Path object as input
    returns Path(dst) on success
    raises FileNotFoundError if src does not exist
    """
    srcPath = pathify(src).resolve()
    dstPath = pathify(dst)
    shcopy(str(srcPath), str(dstPath))
    return dstPath


def copyallfiles(srcdir, dstdir):
    """
    Copies all files in srcdir to dstdir
    """
    srcPath = pathify(srcdir)
    dstPath = pathify(dstdir)

    for p in srcPath.glob('*.*'):
        if p.is_file():
            copy(p, dstPath / p.name)


def touch(filename):
    """
    Creates file if it doesn't exist but always modifies utime.
    Returns a Path(filename)
    """
    path = pathify(filename)
    path.touch(exist_ok=True)
    return path


def ensureflag(flagfile, action=None):
    """Checks if flagfile exist and IF NOT the action function
    will be executed. The flagfile will be 'touched' at the end

    Parameters
    ----------
    flagfile : string
        path to the file used as flag
    action : callable
        this will be called if the flagfiled doesn't exist

    Returns
    -------
    Path(flagfile)
    """
    flagPath = pathify(flagfile)
    if not flagPath.exists() and callable(action):
        action()
    return touch(flagPath)


def ensurefolder(folder):
    """Creates the folder if it doesn't exist

    Parameters
    ----------
    folder : string|pathlib.Path
        path to the folder

    Returns
    -------
    pathlib.Path
        Resolved path of the folder
    """
    folderPath = pathify(folder)
    try:
        makedirs(str(folderPath))
    except FileExistsError:
        pass
    return folderPath.resolve()



from __future__ import absolute_import, with_statement, print_function, division, unicode_literals

from pathlib import Path
from .tradeexcept import TradeException

import csv
import os
import sqlite3

######################################################################
# TradeDangerous :: Modules :: CSV Exporter
#
# Generate a CSV files for a table of the database.
#
# Note: This routine makes some assumptions about the structure
#       of the database:
#          * The table should only have one UNIQUE index
#          * The referenced table must have one UNIQUE index
#          * The FK columns must have the same name in both tables
#          * One column primary keys will be handled by the database engine
#
######################################################################
# CAUTION: If the database structure gets changed this script might
#          need some corrections.
######################################################################

######################################################################
# Default values

# for some tables the first two columns will be reversed
reverseList = []

######################################################################
# Helpers
######################################################################

def search_keyList(list, val):
    for row in list:
        if row['from'] == row['to'] == val: return row

def getUniqueIndex(conn, tableName):
    """ return all unique columns """
    idxCursor = conn.cursor()
    unqIndex = []
    for idxRow in idxCursor.execute("PRAGMA index_list('%s')" % tableName):
        if idxRow['unique']:
            # it's a unique index
            unqCursor = conn.cursor()
            for unqRow in unqCursor.execute("PRAGMA index_info('%s')" % idxRow['name']):
                unqIndex.append(unqRow['name'])
    return unqIndex

def getFKeyList(conn, tableName):
    """ get all single column FKs """
    keyList = []
    keyCount = -1
    keyCursor = conn.cursor()
    for keyRow in keyCursor.execute("PRAGMA foreign_key_list('%s')" % tableName):
        if keyRow['seq'] == 0:
            keyCount += 1
            keyList.append( {'table': keyRow['table'],
                             'from': keyRow['from'],
                             'to': keyRow['to']}
                          )
        if keyRow['seq'] == 1:
            # if there is a second column, remove it from the list
            keyList.remove( keyList[keyCount] )
            keyCount -= 1

    return keyList

def buildFKeyStmt(conn, tableName, key):
    """
        resolve the FK constrain with the UNIQUE index
        multicolumn UNIQUEs are allowed as long as the last one
        will be a single column one
    """
    unqIndex = getUniqueIndex(conn, key['table'])
    keyList  = getFKeyList(conn, key['table'])
    keyStmt = []
    for colName in unqIndex:
        # check if the column is a foreign key
        keyKey = search_keyList(keyList, colName)
        if keyKey:
            newStmt = buildFKeyStmt(conn, key['table'], keyKey)
            for row in newStmt:
                keyStmt.append(row)
        else:
            keyStmt.append({
                'table': tableName,
                'column': colName,
                'joinTable': key['table'],
                'joinColumn': key['to']
            })

    return keyStmt

######################################################################
# Code
######################################################################

def exportTableToFile(tdb, tdenv, tableName, dataPath=None):
    """
        Generate the csv file for tableName in dataPath
        returns lineCount, exportPath
    """

    # path for csv file
    dataPath = dataPath or tdb.dataPath
    if not dataPath.is_dir():
        raise TradeException("Save location '{}' not found.".format(str(dataPath)))

    # connect to the database
    conn = tdb.getDB()
    conn.row_factory = sqlite3.Row

    # prefix for unique/ignore columns
    uniquePfx = "unq:"
    ignorePfx = "!"

    # create CSV files
    exportPath = (dataPath / Path(tableName)).with_suffix(".csv")
    tdenv.DEBUG0("Export Table '{table}' to '{file}'".format(
                    table=tableName, file=str(exportPath)
                    ))

    lineCount = 0
    with exportPath.open("w", encoding='utf-8', newline="\n") as exportFile:
        exportOut = csv.writer(exportFile, delimiter=",", quotechar="'", doublequote=True, quoting=csv.QUOTE_NONNUMERIC, lineterminator="\n")

        cur = conn.cursor()

        # check for single PRIMARY KEY
        pkCount = 0
        for columnRow in cur.execute("PRAGMA table_info('%s')" % tableName):
            # count the columns of the primary key
            if columnRow['pk'] > 0: pkCount += 1

        # build column list
        columnList = []
        for columnRow in cur.execute("PRAGMA table_info('%s')" % tableName):
            # if there is only one PK column, ignore it
            #if columnRow['pk'] > 0 and pkCount == 1: continue
            columnList.append(columnRow)

        if len(columnList) == 0:
            raise TradeException("No columns to export for table '{}'.".format(tableName))

        # reverse the first two columns for some tables
        if tableName in reverseList:
            columnList[0], columnList[1] = columnList[1], columnList[0]

        # initialize helper lists
        csvHead    = []
        stmtColumn = []
        stmtTable  = [ tableName ]
        stmtOrder  = []
        unqIndex   = getUniqueIndex(conn, tableName)
        keyList    = getFKeyList(conn, tableName)

        tdenv.DEBUG1('UNIQUE: ' + ", ".join(unqIndex))

        # iterate over all columns of the table
        for col in columnList:
            # check if the column is a foreign key
            key = search_keyList(keyList, col['name'])
            if key:
                # make the join statement
                keyStmt = buildFKeyStmt(conn, tableName, key)
                for keyRow in keyStmt:
                    tdenv.DEBUG1('FK-Stmt: {}'.format(list(keyRow)))
                    # is the join for the same table
                    if keyRow['table'] == tableName:
                        csvPfx = ''
                        joinStmt = 'USING({})'.format(keyRow['joinColumn'])
                    else:
                        # this column must be ignored by the importer, it's only
                        # used to resolve the FK relation
                        csvPfx = ignorePfx
                        joinStmt = 'ON {}.{} = {}.{}'.format(keyRow['table'], keyRow['joinColumn'], keyRow['joinTable'], keyRow['joinColumn'])
                    if col['name'] in unqIndex:
                        # column is part of an unique index
                        csvPfx = uniquePfx + csvPfx
                    csvHead += [ "{}{}@{}.{}".format(csvPfx, keyRow['column'], keyRow['joinTable'], keyRow['joinColumn']) ]
                    stmtColumn += [ "{}.{}".format(keyRow['joinTable'], keyRow['column']) ]
                    if col['notnull']:
                        stmtTable += [ 'INNER JOIN {} {}'.format(keyRow['joinTable'], joinStmt) ]
                    else:
                        stmtTable += [ 'LEFT OUTER JOIN {} {}'.format(keyRow['joinTable'], joinStmt) ]
                    stmtOrder += [ "{}.{}".format(keyRow['joinTable'], keyRow['column']) ]
            else:
                # ordinary column
                if col['name'] in unqIndex:
                    # column is part of an unique index
                    csvHead += [ uniquePfx + col['name'] ]
                    stmtOrder += [ "{}.{}".format(tableName, col['name']) ]
                else:
                    csvHead += [ col['name'] ]
                stmtColumn += [ "{}.{}".format(tableName, col['name']) ]

        # build the SQL statement
        sqlStmt = "SELECT {} FROM {}".format(",".join(stmtColumn)," ".join(stmtTable))
        if len(stmtOrder) > 0:
            sqlStmt += " ORDER BY {}".format(",".join(stmtOrder))
        tdenv.DEBUG1("SQL: %s" % sqlStmt)

        # finally generate the csv file
        # write header line without quotes
        exportFile.write("{}\n".format(",".join(csvHead)))
        for line in cur.execute(sqlStmt):
            lineCount += 1
            tdenv.DEBUG2("{count}: {values}".format(count=lineCount, values=list(line)))
            exportOut.writerow(list(line))
        tdenv.DEBUG1("{count} {table}s exported".format(count=lineCount, table=tableName))

    # Update the DB file so we don't regenerate it.
    os.utime(str(tdb.dbPath))

    return lineCount, exportPath



class TradeException(Exception):
    """
        Distinguishes runtime logical errors (such as no data for what you
        queried) from programmatic errors (such as Oliver accessing a hash
        with the wrong type of key).

        TradeExcepts should be caught by the program and displayed in the
        most user friendly way possible.
    """
    pass


#! /usr/bin/env python

import pathlib
import re
import sys

try:
    import requests
except ImportError as e:
    print("""ERROR: Unable to load the Python 'requests' package.

This script uses a Python module/package called 'requests' to allow
it to talk to maddavo's web service. This package is not installed
by default, but it can be installed with Python's package manager (pip).

You can either install/update it yourself, e.g.:

  pip install --upgrade requests

or if you like, I can try and install it for you now
""")
    approval = input(
        "Do you want me to try and install it with the package manager (y/n)? "
    )
    if approval.lower() != 'y':
        print("You didn't type 'y' so I'm giving up.")
        raise e
    import pip
    pip.main(["install", "--upgrade", "requests"])
    import requests


############################################################################


upload_url = 'http://www.davek.com.au/td/uploaddata.asp'
if len(sys.argv) != 2:
    raise SystemExit(
        "Usage: {} <filename>\n"
        "Uploads 'filename' to Maddavo's site, where filename should be "
        "a .prices, Station.csv or ShipVendor.csv file.\n"
        "\n"
        "NOTE to OCR USERS:\n"
        "Please upload the .prices file from your OCR DIRECTLY before "
        "importing it into TD, not after.".format(sys.argv[0])
    )

upfile = sys.argv[1]
uppath = pathlib.Path(upfile)

############################################################################

if not uppath.exists():
    raise SystemExit("ERROR: {}: File not found".format(upfile))
if not uppath.is_file():
    raise SystemExit("ERROR: {}: Not a file".format(upfile))

r = requests.post(
        upload_url,
        files={
            'Filename': (
                uppath.name,
                uppath.open("rb"),
                'text/plain',
                {
                    "Expires": '90',
                }
            ),
        }
)

response = r.text
m = re.search(r'UPLOAD RESULT:\s*(.*?)<br', response, re.IGNORECASE)
if not m:
    raise Exception("Unexpected result:\n" + r.text)

resultCode = m.group(1)
if resultCode.startswith("SUCCESS"):
    raise SystemExit("Upload complete")

print("Error response from Maddavo's site: {}".format(resultCode))
with open("tmp/maderror.txt", "w", encoding="utf-8") as out:
    print(r.text, file=out)
    print("See tmp/maderror.txt for full response.")


import sys

class Progress(object):
    """
    Helper class that describes a simple text-based progress bar.
    """

    def __init__(self, maxValue, width, start=0, prefix=""):
        """
        Arguments:
            maxValue
                Last value we can reach (100%).
            width
                Number of '='s characters for 100%
            start
                Initial value
            prefix
                Something to print infront of the progress bar
        """
        self.start = start
        self.maxValue = maxValue
        self.width = width
        self.value = start
        self.progress = -1
        self.prefix = prefix
        self.textLen = 0
        self.mask = '\r' + self.prefix + "[{{:<{width}}}]".format(width=width)


    def increment(self, value, postfix=""):
        """
        Increment the progress bar's internal counter by 'value',
        and if this changes the progress step, re-draw the bar.

        Attributes:
            value
                The amount to increment the internal counter by
            postfix [optional]
                String or callable to print after the bar

        Returns:
            False if the progress bar did not redraw,
            True if the progress bar was redrawn,
        """
        self.value = min(self.maxValue, self.value + value)
        progress = int(self.width * (self.value - self.start) / self.maxValue)
        if progress == self.progress:
            return False

        if callable(postfix):
            postfixText = postfix(self.value, self.maxValue)
        else:
            postfixText = postfix
        text = self.mask.format("="*progress) + postfixText + " "
        sys.stdout.write(text)
        sys.stdout.flush()

        self.progress = progress
        self.textLen = len(text)

        return True


    def clear(self):
        """
        Remove the current progress bar, if any
        """
        if self.textLen:
            fin = "\r{:{width}}\r".format('', width=self.textLen)
            sys.stdout.write(fin)
            sys.stdout.flush()


from __future__ import absolute_import, with_statement, print_function, division, unicode_literals

from ..csvexport import exportTableToFile
from .parsing import *
from .exceptions import CommandLineError
from pathlib import Path

import sqlite3

######################################################################
# TradeDangerous :: Commands :: Export
#
# Generate the CSV files for the master data of the database.
#
######################################################################
# CAUTION: If the database structure gets changed this script might
#          need some corrections.
######################################################################

######################################################################
# Parser config

help='CSV exporter for TradeDangerous database.'
name='export'
epilog=(
        "CAUTION: If you don't specify a different path, the current "
        "CSV files in the data directory will be overwritten with "
        "the current content of the database.\n "
        "If you have changed any CSV file and didn't rebuild the "
        "database, they will be lost.\n "
        "Use the 'buildcache' command first to rebuild the database."
)
wantsTradeDB=False           # because we don't want the DB to be rebuild
arguments = [
]
switches = [
    ParseArgument('--path',
            help="Specify a different save location of the CSV files than the default.",
            type=str,
            default=None
        ),
    MutuallyExclusiveGroup(
        ParseArgument('--tables', "-T",
                help='Specify comma separated tablenames to export.',
                metavar='TABLE[,TABLE,...]',
                type=str,
                default=None
            ),
        ParseArgument('--all-tables',
                help='Include the price tables for export.',
                dest='allTables',
                action='store_true',
                default=False
            ),
        ),
    ParseArgument('--delete-empty',
            help='Delete CSV files without content.',
            dest='deleteEmpty',
            action='store_true',
            default=False
        ),
]

######################################################################
# Perform query and populate result set

def run(results, cmdenv, tdb):
    # check database exists
    if not tdb.dbPath.is_file():
        raise CommandLineError("Database '{}' not found.".format(tdb.dbPath))

    # check export path exists
    if cmdenv.path:
        # the "--path" overwrites the default path of TD
        exportPath = Path(cmdenv.path)
    else:
        exportPath = Path(cmdenv.dataDir)
    if not exportPath.is_dir():
        raise CommandLineError("Save location '{}' not found.".format(str(exportPath)))

    # connect to the database
    cmdenv.NOTE("Using database '{}'", tdb.dbPath)
    conn = tdb.getDB()
    conn.row_factory = sqlite3.Row

    # some tables might be ignored
    ignoreList = []

    # extract tables from command line
    if cmdenv.tables:
        bindValues = cmdenv.tables.split(',')
        tableStmt = " AND name COLLATE NOCASE IN ({})".format(",".join("?" * len(bindValues)))
        cmdenv.DEBUG0(tableStmt)
    else:
        bindValues = []
        tableStmt = ''
        if not cmdenv.allTables:
            ignoreList.append("StationItem")

    tableCursor = conn.cursor()
    for row in tableCursor.execute("""
                                      SELECT name
                                        FROM sqlite_master
                                       WHERE type = 'table'
                                         AND name NOT LIKE 'sqlite_%'
                                             {cmdTables}
                                       ORDER BY name
                                   """.format(cmdTables=tableStmt),
                                   bindValues):
        tableName = row['name']
        if tableName in ignoreList:
            # ignore the table
            cmdenv.NOTE("Ignore Table '{table}'", table=tableName)
            continue

        cmdenv.NOTE("Export Table '{table}'", table=tableName)

        # create CSV files
        lineCount, filePath = exportTableToFile(tdb, cmdenv, tableName, exportPath)
        if cmdenv.deleteEmpty and lineCount == 0:
            # delete file if emtpy
            filePath.unlink()
            cmdenv.DEBUG0("Delete empty file {file}'".format(file=filePath))

    return None


#!/usr/bin/env python3
# --------------------------------------------------------------------
# Copyright (C) Oliver 'kfsone' Smith 2014 <oliver@kfs.org>:
# Copyright (C) Bernd 'Gazelle' Gollesch 2016, 2017
# Copyright (C) Jonathan 'eyeonus' Jones 2018
#
# You are free to use, redistribute, or even print and eat a copy of
# this software so long as you include this copyright notice.
# I guarantee there is at least one bug neither of us knew about.
# --------------------------------------------------------------------

"""just keeper of current version"""

#TODO: remember to update tests when version changes
__version__ = '10.0.2'




from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
import itertools

class ColumnFormat(object):
    """
        Describes formatting of a column to be populated with data.

        Member Functions:

            str()
                Applies all formatting (except qualifier) to the name to
                produce a correctly sized title field.

            format(value)
                Applies all formatting to key(value) to produce a correctly
                sized value field.

        Attributes:
            name
                Heading for column to display when calling title()
                e.g. name="Station Name"
            align
                Alignment formatter for .format
                e.g. align='<' or align='>' or align=''
            width
                Numeric value for the width of the column
                e.g. width=5
            qualifier
                Final part of the print format,
                e.g. qualifier='.2f' or qualifier='n'
            pre
                Prefix to the column
            post
                Postfix to the column
            key
                Retrieve the printable name of the item
            pred
                Predicate: Return False to leave this column blank

        e.g.
            cols = [
                ColumnFormat("Name", "<", "5", '', key=lambda item:item['name']),
                ColumnFormat("Dist", ">",  "6", ".2f",
                        pre='[',
                        post=']'
                        key=lambda item:item['dist']),
            ]
            rows = [ {'name':'Bob', 'dist':1.5}, {'name':'John', 'dist':23}]
            # print titles
            print(*[col.str() for col in cols])
            for row in rows:
                print(*[col.format(row) for col in cols])
        Produces:
            Name   [ Dist]
            Bob    [ 1.30]
            John   [23.00]

    """
    def __init__(
            self,
            name,
            align,
            width,
            qualifier=None,
            pre=None,
            post=None,
            key=lambda item: item,
            pred=lambda item: True,
            ):
        self.name = name
        self.align = align
        self.width = max(int(width), len(name))
        self.qualifier = qualifier or ''
        self.key = key
        self.pre = pre or ''
        self.post = post or ''
        self.pred = pred


    def str(self):
        return '{pre}{title:{align}{width}}{post}'.format(
                title=self.name,
                align=self.align, width=self.width,
                pre=self.pre, post=self.post,
            )


    def format(self, value):
        if self.pred(value):
            return '{pre}{value:{align}{width}{qual}}{post}'.format(
                    value=self.key(value),
                    align=self.align, width=self.width,
                    qual=self.qualifier,
                    pre=self.pre, post=self.post,
                )
        else:
            return '{pre}{value:{align}{width}}{post}'.format(
                value="",
                align=self.align, width=self.width,
                pre=self.pre, post=self.post,
            )


class RowFormat(object):
    """
        Describes an ordered collection of ColumnFormats
        for dispay data from rows, such that calling
          rowFmt.format(rowData)
        will return the result of formatting each column
        against rowData.

        Member Functions

            append(col, [after])
                Adds a ColumnFormatter to the end of the row
                If 'after' is specified, tries to insert
                the new column immediately after the first
                column who's name matches after.

            insert(pos, newCol)
                Inserts a ColumnFormatter at position pos in the list

            str()
                Returns a list of all the column headings

            format(rowData):
                Returns a list of applying rowData to all
                of the columns

    """
    def __init__(self, prefix=None):
        self.columns = []
        self.prefix = prefix or ""


    def addColumn(self, *args, **kwargs):
        self.append(ColumnFormat(*args, **kwargs))


    def append(self, column, after=None):
        columns = self.columns
        if after:
            for idx, col in enumerate(columns, 1):
                if col.name == after:
                    columns.insert(idx, column)
                    return self
        columns.append(column)
        return self


    def insert(self, pos, column):
        if column is not None:
            self.columns.insert(pos, column)


    def str(self):
        return self.prefix + ' '.join(col.str() for col in self.columns)


    def heading(self):
        headline = self.str()
        return headline, '-' * len(headline)


    def format(self, rowData):
        return self.prefix + ' '.join(col.format(rowData) for col in self.columns)

def max_len(iterable, key=lambda item: item):
    iterable, readahead = itertools.tee(iter(iterable))
    try:
        next(readahead)
    except StopIteration:
        return 0
    return max(len(key(item)) for item in iterable)

if __name__ == '__main__':
    rowFmt = RowFormat(). \
                append(ColumnFormat("Name", '<', '8', key=lambda row: row['name'])). \
                append(ColumnFormat("Dist", '>', '6', '.2f', pre='[', post=']', key=lambda row: row['dist']))

    rows = [
        { 'name': 'Bob', 'dist': 6.2, 'age': 30 },
        { 'name': 'Dave', 'dist': 42, 'age': 18 },
    ]

    def present():
        rowTitle = rowFmt.str()
        print(rowTitle)
        print('-' * len(rowTitle))
        for row in rows:
            print(rowFmt.format(row))

    print("Simple usage:")
    present()

    print()
    print("Adding age ColumnFormat:")

    rowFmt.append(after='Name', col=ColumnFormat("Age", '>', 3, pre='|', post='|', key=lambda row: row['age']))
    present()


import csvexport
import misc.eddb
import tradedb

# Build an ID=>Name mapping for EDDB systems
tdb = tradedb.TradeDB()
systems = {}
tdSysLookup = tdb.systemByName.get
for s in misc.eddb.SystemsQuery():
    tdSys = tdSysLookup(s['name'].upper(), None)
    if tdSys and tdSys.stations:
        systems[s['id']] = { stn.dbname.upper(): stn for stn in tdSys.stations }

def matching_stations():
    # generator that searches the eddb station set for entries that
    # match tdb entries and yields the tuple (tdStn, eddbStn)

    for eddbStn in misc.eddb.StationsQuery():
        stationList = systems.get(eddbStn['system_id'], None)
        if not stationList:
            continue
        name = eddbStn['name'].upper()
        tdStn = stationList.get(name, None)
        if tdStn:
            yield tdStn, eddbStn

updateStation = tdb.updateLocalStation

bool_trans = { None: '?', 0: 'N', 1: 'Y' }

updates = 0
for tdStn, eddbStn in matching_stations():
    mps = eddbStn['max_landing_pad_size'] or '?'
    if updateStation(
            station=tdStn,
            lsFromStar=eddbStn['distance_to_star'],
            maxPadSize=mps,
            market=bool_trans[eddbStn['has_commodities']],
            blackMarket=bool_trans[eddbStn['has_blackmarket']],
            shipyard=bool_trans[eddbStn['has_shipyard']],
            outfitting=bool_trans[eddbStn['has_outfitting']],
            rearm=bool_trans[eddbStn['has_rearm']],
            refuel=bool_trans[eddbStn['has_refuel']],
            repair=bool_trans[eddbStn['has_repair']],
            modified='now',
            commit=False,
            ):
        updates += 1

if updates:
    tdb.getDB().commit()
    csvexport.exportTableToFile(tdb, tdb.tdenv, "Station")
    print("Updated Station.csv: {} updates".format(updates))


#!/usr/bin/env python3.6

"""
Small tool to submit new star data to EDStarCoordinator.

Use:

    submit-distances.py "system name"
"""

#
# NOTE: This tool is very hacky. If someone wants to clean it up,
# and submit a diff, that'd be greatly appreciated!
#

from __future__ import print_function

import argparse
import json
import math
import os
import pathlib
import platform
import random
import re
import sys
import tradedb
import tradeenv

from misc.edsc import StarSubmission, StarSubmissionResult, SubmissionError
from misc.clipboard import SystemNameClip

try:
    import requests
except ImportError as e:
    print("""ERROR: Unable to load the Python 'requests' package.

This script uses a Python module/package called 'requests' to allow
it to talk to the EDSC web service. This package is not installed
by default, but it can be installed with Python's package manager (pip).

You can either install/update it yourself, e.g.:

  pip install --upgrade requests

or if you like, I can try and install it for you now
""")
    approval = input(
        "Do you want me to try and install it with the package manager (y/n)? "
    )
    if approval.lower() != 'y':
        print("You didn't type 'y' so I'm giving up.")
        raise e
    import pip
    pip.main(["install", "--upgrade", "requests"])
    import requests

standardStars = [
    "SOL",
    "NEW YEMBO",
    "VESUVIT",
    "HIP 79884",
    "ASELLUS AUSTRALIS",
]

############################################################################


class UsageError(Exception):
    def __init__(self, argv, error):
       self.argv, self.error = argv, error
    def __str__(self):
        return error + "\n" + argv.format_usage()


def parse_arguments():
    parser = argparse.ArgumentParser(
        description='Submit star distances to the EDSC project.',
    )
    parser.add_argument(
        'origin',
        help='System to submit distances for.',
        type=str,
    )
    parser.add_argument(
        '--pick',
        help='Randomly select N existing systems from gamma or ealier',
        required=False,
        type=int,
    )
    parser.add_argument(
        '--cmdr',
        required=False,
        help='Specify your commander name.',
        type=str,
        default=os.environ.get('CMDR', None)
    )
    parser.add_argument(
        '--test',
        help='Use the EDSC Test Database.',
        required=False,
        action='store_true',
    )
    parser.add_argument(
        '--detail', '-v',
        help='Output additional detail.',
        action='count',
        default=0,
    )
    parser.add_argument(
        '--debug', '-w',
        help='Enable debugging output,',
        action='count',
        default=0,
    )
    parser.add_argument(
        '--no-update',
        help='Disallow distances from an existing system.',
        action='store_false',
        default=True,
        dest='allowUpdate',
    )
    parser.add_argument(
        '--extra-file',
        help='File to read/write extra stars to.',
        type=str,
        default='data/extra-stars.txt',
        dest='extraFile',
    )
    parser.add_argument(
        'destinations',
        help='System or systems to measure distance to.',
        default=[],
        nargs='*',
    )

    argv = parser.parse_args(sys.argv[1:])
    argv.origin = argv.origin.upper()
    if argv.origin.startswith('@'):
        argv.origin = argv.origin[1:]

    if not argv.cmdr:
        raise UsageError(argv, "No commander name specified")

    return argv


def get_system(argv, tdb):
    system = tdb.systemByName.get(argv.origin.upper(), None)
    if not system:
        return argv.origin, None

    if not argv.allowUpdate:
        raise UsageError(
            argv,
            "System '{}' already exists.\n"
            .format(systemName)
        )

    if argv.detail:
        print("EXISTING SYSTEM:", argv.origin)

    return argv.origin, system


def pick_destinations(argv, tdb):
    numSystems = len(tdb.systemByName)
    if numSystems < 1:
        raise UsageError(
            argv,
            "Can't --pick random systems: "
            "Your TD database doesn't contain any systems."
        )
    num = min(argv.pick, numSystems)
    systems = tdb.systemByName
    try:
        gamma = tdb.lookupAdded("Gamma")
    except KeyError:
        gamma = 20
    destinations = random.sample([
        sysName for sysName, system in systems.items()
        if system.addedID <= gamma and \
            not sysName in standardStars
    ], num)

    return destinations


def get_outliers(argv):
    outliers = set()
    try:
        with open(argv.extraFile, "rU", encoding="utf-8") as input:
            for line in input:
                name = line.partition('#')[0].strip().upper()
                if name and name != argv.origin:
                    outliers.add(name)
    except FileNotFoundError:
        pass
    outliers = list(outliers)
    random.shuffle(outliers)
    return outliers


def get_distances(argv, clip, stars):
    distances = []
    for star in stars:
        starNo = len(distances) + 1
        # Check it's not already in the list
        star = star.upper()
        if star in distances:
            continue

        clip.copy_text(star)

        if len(stars) > 1:
            prefix = "{:>2}/{:2}: ".format(starNo, len(stars))
        else:
            prefix = ""
        dist = input(prefix + "Distance to {}: ".format(star))
        if dist == 'q':
            return distances, 'q'

        if dist:
            distances.append({
                'name': star.upper(),
                'dist': float(dist),
            })

    return distances, 'end'
 

def check_system(argv, tdb, tdbSys, name):
    system = tdb.systemByName.get(name.upper(), None)
    if system and not tdbSys:
        print("KNOWN SYSTEM")
        return
    if system:
        print("KNOWN SYSTEM: {:.2f} ly".format(
            tdbSys.distanceTo(system)
        ))


def add_extra_stars(argv, extraStars):
    if not extraStars:
        return
    if argv.detail:
        print("Saving {} to {}".format(
            str(extraStars), argv.extraFile,
        ))
    try:
        with open(argv.extraFile, "a", encoding="utf-8") as output:
            for star in extraStars:
                print(star, file=output)
    except FileNotFoundError:
        pass


def submit_distances(argv, tdb, distances):
    system = argv.origin
    cmdr = argv.cmdr
    mode = "TEST" if argv.test else "Live"

    print()
    print("System:", system)
    print("Database:", mode)
    print("Distances:")
    for ref in distances:
        print("  {}: {:.02f} ly".format(
            ref['name'], ref['dist']
        ))
    print()

    ok = input("Does this look correct (y/n)? ")
    if ok != 'y':
        print("Stopped")
        return

    print()
    print("Submitting {} {}".format(mode, system))

    sub = StarSubmission(
        star=system,
        commander=cmdr,
        refs=distances,
        test=argv.test,
    )
    resp = sub.submit()

    result = StarSubmissionResult(star=system, response=resp)
    print(str(result))
    if result.valid:
        # Check for systems we can add
        trilats = set()
        for sysName in result.systems.keys():
            code, coord = result.systems[sysName]
            sysName = sysName.upper()
            # Does it have a distance
            if isinstance(coord, (list, tuple)):
                x, y, z = coord
                system = tdb.systemByName.get(sysName, None)
                if system:
                    tdb.updateLocalSystem(system, sysName, x, y, z)
                else:
                    tdb.addLocalSystem(sysName, x, y, z)
        if result.recheck:
            return list(result.recheck.keys())
    return None


def do_rechecks(argv, clip, rechecks):
    print("\aSome systems need their distances rechecked:")

    distances, term = get_distances(argv, clip, rechecks)
    return distances


def send_and_check_distances(argv, tdb, clip, distances):
    if not distances:
        if argv.detail:
            print("No distances, no submission.")
        return False

    while distances:
        rechecks = submit_distances(argv, tdb, distances)
        if not rechecks:
            break

        distances = do_rechecks(argv, clip, rechecks)

    return True


def process_destinations(argv, tdb):
    clip = SystemNameClip()

    print("Distances from {}:".format(argv.origin))
    distances, _ = get_distances(argv, clip, argv.destinations)
    send_and_check_distances(argv, tdb, clip, distances)


############################################################################

def main():
    argv = parse_arguments()

    tdenv = tradeenv.TradeEnv(properties=argv)
    tdb = tradedb.TradeDB(tdenv)

    system, tdbSys = get_system(argv, tdb)

    if argv.pick:
        argv.destinations.extend(pick_destinations(argv, tdb))

    if argv.destinations:
        process_destinations(argv, tdb)
        return

    print("Add EDSC Star Distances for \"{}\"".format(system))
    print()
    print("You will now be prompted for distances to various stars.")
    print()
    print(
        "At each prompt, the star name will be copied into your paste buffer. "
        "You should alt-tab into the game and paste the name into the Galaxy "
        "Map's search box. Then alt-tab back and enter the distance value."
    )
    print()
    print(
        "At each prompt enter a ly distance (e.g. 123.45), q to stop, "
        "or leave the line empty if you don't want to enter data for "
        "this star."
    )

    print(
        "5 distances are required for EDSC to make a first guess at a "
        "star's location. You can submit more to increase the accuracy "
        "but the only time you need to submit more than 10 is when you "
        "are trying to submit corrections."
    )
    print()

    clip = SystemNameClip()

    print()
    print("""
===================================================
STANDARD STARS: (q to skip to the next section)

  These are stars with well-known positions.
===================================================
""")
    distances, term = get_distances(argv, clip, standardStars)
    if distances:
        send_and_check_distances(argv, tdb, clip, distances)

    outliers = get_outliers(argv)
    if outliers:
        print("""
===================================================
EXTRA STARS: (q to skip to the next section)

  Stars from {}.
===================================================
""".format(argv.extraFile))
        distances, term = get_distances(argv, clip, outliers)
        if distances:
            send_and_check_distances(argv, tdb, clip, distances)

    print("""
===================================================
CHOOSE YOUR OWN: (q to stop)

  Specify additional stars.

  Prefix names with a '+' to add them to
  {}.
===================================================
""".format(argv.extraFile))
    distances = []
    newOutliers = []
    while True:
        star = input("Enter star name: ")
        star = star.strip().upper()
        if not star or star == 'Q':
            break
        # Remove surrounding quotes
        save = False
        if star.startswith('+'):
            save = True
            star = star[1:].strip()
        star = re.sub(r'\s+', ' ', star)
        star = re.sub(r"''+", "'", star)
        star = re.sub(r'^("|\')+\s*(.*)\s*\1+', r'\2', star)
        if star.find('"') >= 0:
            print("Invalid star name")
            continue
        if star.startswith('+'):
            save = True
            star = star[1:].strip()
        for ref in distances:
            if ref['name'] == star:
                print("'{}' is already listed.")
                continue
        check_system(argv, tdb, tdbSys, star)
        extras, term = get_distances(argv, clip, [star])
        if term != 'q' and len(extras) > 0:
            distances.extend(extras)
            if save and star not in outliers and star not in newOutliers:
                newOutliers.append(star)

    if send_and_check_distances(argv, tdb, clip, distances):
        add_extra_stars(argv, newOutliers)

if __name__ == "__main__":
    try:
        main()
    except (SubmissionError, UsageError) as e:
        print(str(e))




"""
DirectOutput.py - Saitek DirectOutput.dll Python Wrapper

Version: 0.3
Author: Oliver "kfsone" Smith <oliver@kfs.org> 2014
Original Author: Frazzle

Description: Python wrapper class for DirectOutput functions.

This module consists of two classes - DirectOutput and DirectOutputDevice

DirectOutput directly calls C functions within DirectOutput.dll to allow Python control of the Saitek X52 Pro MFD and LEDs. Implemented as a class to allow sharing of dll object amongst functions

DirectOutputDevice is a wrapper around DirectOutput which automates setup and persists the device handle across functions. This class can be directly called or inherited to control an individual device (eg. X52 Pro)

Thanks to Spksh and ellF for the C# version of the wrapper which was very helpful in implementing this.
Thanks to Frazzle for the first Python version (no-longer compatible with the saitek driver).

Example Usage:

device = DirectOutputDevice()
device.AddPage(0, "Test", True)
device.SetString(0, 0, "Test String")

import time, sys
while True:
    try:
        time.sleep(1)
    except:
        #This is used to catch Ctrl+C, calling finish method is *very* important to de-initalize device.
        device.finish()
        sys.exit()

Saitek appear to periodically change the DLLs functions. At time of writing,
these are the functions listed in the DLL:

DirectOutput_Initialize
DirectOutput_Deinitialize
DirectOutput_AddPage
DirectOutput_DeleteFile
DirectOutput_DisplayFile
DirectOutput_Enumerate
DirectOutput_GetDeviceInstance
DirectOutput_GetDeviceType
DirectOutput_RegisterDeviceCallback
DirectOutput_RegisterPageCallback
DirectOutput_RegisterSoftButtonCallback
DirectOutput_RemovePage
DirectOutput_SaveFile
DirectOutput_SendServerFile
DirectOutput_SendServerMsg
DirectOutput_SetImage
DirectOutput_SetImageFromFile
DirectOutput_SetLed
DirectOutput_SetProfile
DirectOutput_SetString

DirectOutput_StartServer
DirectOutput_CloseServer

"""

from __future__ import absolute_import, with_statement, print_function, division, unicode_literals

from mfd import MissingDeviceError

import ctypes
import ctypes.wintypes
import logging
import os
import platform

S_OK = 0x00000000
E_HANDLE = 0x80070006
E_INVALIDARG = 0x80070057
E_OUTOFMEMORY = 0x8007000E
E_PAGENOTACTIVE = -0xfbffff        # Something munges it from it's actual value
E_BUFFERTOOSMALL = -0xfc0000
E_NOTIMPL = 0x80004001
ERROR_DEV_NOT_EXIST = 55

SOFTBUTTON_SELECT = 0x00000001
SOFTBUTTON_UP = 0x00000002
SOFTBUTTON_DOWN = 0x00000004


class DirectOutput(object):

    def __init__(self, dll_path):
        """
        Creates python object to interact with DirecOutput.dll

        Required Arguments:
        dll_path -- String containing DirectOutput.dll location.

        """
        logging.debug("DirectOutput.__init__")
        self.DirectOutputDLL = ctypes.WinDLL(dll_path)

    def Initialize(self, application_name):
        """
        Function to call DirectOutput_Initialize

        Required Arguments:
        application_name -- String representing name of applicaiton - must be unique per-application

        Returns:
        S_OK: The call completed sucesfully
        E_OUTOFMEMORY: There was insufficient memory to complete this call.
        E_INVALIDARG: The argument is invalid
        E_HANDLE: The DirectOutputManager process could not be found

        """
        logging.debug("DirectOutput.Initialize")
        return self.DirectOutputDLL.DirectOutput_Initialize(ctypes.wintypes.LPWSTR(application_name))

    def Deinitialize(self):
        """
        Direct function call to DirectOutput_Deinitialize

        Returns:
        S_OK: The call completed successfully.
        E_HANDLE:  DirectOutput was not initialized or was already deinitialized.
        """
        logging.debug("DirectOutput.Deinitialize")
        return self.DirectOutputDLL.DirectOutput_Deinitialize()

    def RegisterDeviceCallback(self, function):
        """
        Registers self.DeviceCallback to be called when devices get registered

        Required Arugments:
        function -- Function to call when a device registers

        Returns:
        S_OK: The call completed successfully
        E_HANDLE: DirectOutput was not initialized.

        """
        logging.debug("DirectOutput.RegisterDeviceCallback")
        return self.DirectOutputDLL.DirectOutput_RegisterDeviceCallback(function, 0)

    def Enumerate(self, function):
        """
        Direct call to DirectOutput_Enumerate

        Returns:
        S_OK: The call completed successfully
        E_HANDLE: DirectOutput was not initialized.

        """
        logging.debug("DirectOutput.Enumerate")
        return self.DirectOutputDLL.DirectOutput_Enumerate(function, 0)


    def RegisterSoftButtonCallback(self, device_handle, function):
        """
        Registers a function to be called when a soft button changes

        Required Arugments:
        device_handle -- ID of device
        function -- Function to call when a soft button changes

        Returns:
        S_OK: The call completed successfully.
        E_HANDLE: The device handle specified is invalid.

        """
        logging.debug("DirectOutput.RegisterSoftButtonCallback({}, {})".format(device_handle, function))
        return self.DirectOutputDLL.DirectOutput_RegisterSoftButtonCallback(device_handle, function, 0)

    def RegisterPageCallback(self, device_handle, function):
        """
        Registers a function to be called when page changes

        Required Arugments:
        device_handle -- ID of device
        function -- Function to call when a page changes

        Returns:
        S_OK: The call completed successfully.
        E_HANDLE: The device handle specified is invalid.
        """
        logging.debug("DirectOutput.RegisterPageCallback({}, {})".format(device_handle,function))
        return self.DirectOutputDLL.DirectOutput_RegisterPageCallback(device_handle, function, 0)

    def SetProfile(self, device_handle, profile):
        """
        Sets the profile used on the device.

        Required Arguments:
        device_handle -- ID of device
        profile -- full path of the profile to activate. passing None will clear the profile.
        """
        logging.debug("DirectOutput.SetProfile({}, {})".format(device_handle, profile))
        if profile:
            return self.DirectOutputDLL.DirectOutput_SetProfile(device_handle, len(profile), ctypes.wintypes.LPWSTR(profile))
        else:
            return self.DirectOutputDLL.DirectOutput_SetProfile(device_handle, 0, 0)

    def AddPage(self, device_handle, page, name, active):
        """
        Adds a page to the MFD

        Required Arguments:
        device_handle -- ID of device
        page -- page ID to add
        name -- String specifying page name
        active -- True if page is to become the active page, if False this will not change the active page

        Returns:
        S_OK: The call completed successfully.
        E_OUTOFMEMORY: Insufficient memory to complete the request.
        E_INVALIDARG: The dwPage parameter already exists.
        E_HANDLE: The device handle specified is invalid.

        """
        logging.debug("DirectOutput.AddPage({}, {}, {}, {})".format(device_handle, page, name, active))
        return self.DirectOutputDLL.DirectOutput_AddPage(device_handle, page, active)

    def RemovePage(self, device_handle, page):
        """
        Removes a page from the MFD

        Required Arguments:
        device_handle -- ID of device
        page -- page ID to remove

        Returns:
        S_OK: The call completed successfully.
        E_INVALIDARG: The dwPage argument does not reference a valid page id.
        E_HANDLE: The device handle specified is invalid.

        """
        logging.debug("DirectOutput.RemovePage({}, {})".format(device_handle, page))
        return self.DirectOutputDLL.DirectOutput_RemovePage(device_handle, page)

    def SetLed(self, device_handle, page, led, value):
        """
        Sets LED state on a given page

        Required Arguments:
        device_handle -- ID of device
        page -- page number
        led -- ID of LED to change
        value -- value to set LED (1 = on, 0 = off)

        Returns:
        S_OK: The call completes successfully.
        E_INVALIDARG: The dwPage argument does not reference a valid page id, or the dwLed argument does not specifiy a valid LED id.
        E_HANDLE: The device handle specified is invalid

        """
        logging.debug("DirectOutput.SetLed({}, {}, {}, {})".format(device_handle, page, led, value))
        return self.DirectOutputDLL.DirectOutput_SetLed(device_handle, page, led, value)


    def SetString(self, device_handle, page, line, string):
        """
        Sets a string to display on the MFD

        Required Arguments:
        device_handle -- ID of device
        page -- the ID of the page to add the string to
        line -- the line to display the string on (0 = top, 1 = middle, 2 = bottom)
        string -- the string to display

        Returns:
        S_OK: The call completes successfully.
        E_INVALIDARG: The dwPage argument does not reference a valid page id, or the dwString argument does not reference a valid string id.
        E_OUTOFMEMORY: Insufficient memory to complete the request.
        E_HANDLE: The device handle specified is invalid.

        """
        logging.debug("DirectOutput.SetString({}, {}, {}, {})".format(device_handle, page, line, string))
        return self.DirectOutputDLL.DirectOutput_SetString(device_handle, page, line, len(string), ctypes.wintypes.LPWSTR(string))


class DirectOutputDevice(object):

    class Buttons(object):

        select, up, down = False, False, False

        def __init__(self, bitmask):
            self.bitmask = bitmask
            if bitmask == 1:
                self.select = True
            elif bitmask == 2:
                self.up = True
            elif bitmask == 3:
                self.up = True
                self.select = True
            elif bitmask == 4:
                self.down = True
            elif bitmask == 5:
                self.down = True
                self.select = True
            elif bitmask == 6:
                self.up = True
                self.down = True
            elif bitmask == 7:
                self.up = True
                self.down = True
                self.select = True

        def __repr__(self):
            return "Select: "+str(self.select)+" Up: "+str(self.up)+" Down: "+str(self.down)

    application_name = "GenericDevice"
    device_handle = None
    direct_output = None
    debug_level = 0

    def __init__(self, debug_level=0, name=None):
        """
        Initialises device, creates internal state (device_handle) and registers callbacks.

        """

        logging.info("DirectOutputDevice.__init__")

        prog_dir = os.environ["ProgramFiles"]
        if platform.machine().endswith('86'):
            # 32-bit machine, nothing to worry about
            pass
        elif platform.machine().endswith('64'):
            # 64-bit machine, are we a 32-bit python?
            if platform.architecture()[0] == '32bit':
                prog_dir = os.environ["ProgramFiles(x86)"]
        dll_path = os.path.join(prog_dir, "Saitek\\DirectOutput\\DirectOutput.dll")

        self.application_name = name or DirectOutputDevice.application_name
        self.debug_level = debug_level

        try:
            logging.debug("DirectOutputDevice -> DirectOutput: {}".format(dll_path))
            self.direct_output = DirectOutput(dll_path)
            logging.debug("direct_output = {}".format(self.direct_output))
        except WindowsError as e:
            logging.warning("DLLError: {}: {}".format(dll_path, e.winerror))
            raise DLLError(e.winerror) from None

        result = self.direct_output.Initialize(self.application_name)
        if result != S_OK:
            logging.warning("direct_output.Initialize returned {}".format(result))
            raise DirectOutputError(result)

        logging.info("Creating callback closures.")
        self.onDevice_closure = self._OnDeviceClosure()
        logging.debug("onDevice_closure is {}".format(self.onDevice_closure))
        self.onEnumerate_closure = self._OnEnumerateClosure()
        logging.debug("onEnumerate_closure is {}".format(self.onEnumerate_closure))
        self.onPage_closure = self._OnPageClosure()
        logging.debug("onPage_closure is {}".format(self.onPage_closure))
        self.onSoftButton_closure = self._OnSoftButtonClosure()
        logging.debug("onSoftButton_closure is {}".format(self.onSoftButton_closure))

        result = self.direct_output.RegisterDeviceCallback(self.onDevice_closure)
        if result != S_OK:
            logging.warning("RegisterDeviceCallback failed: {}".format(result))
            self.finish()
            raise DirectOutputError(result)

        result = self.direct_output.Enumerate(self.onEnumerate_closure)
        if result != S_OK:
            logging.warning("Enumerate failed: {}".format(result))
            self.finish()
            raise DirectOutputError(result)

        if not self.device_handle:
            logging.warning("No device handle")
            self.finish()
            raise MissingDeviceError()

        result = self.direct_output.RegisterSoftButtonCallback(self.device_handle, self.onSoftButton_closure)
        if result != S_OK:
            logging.warning("RegisterSoftButtonCallback failed")
            self.finish()
            raise DirectOutputError(result)

        result = self.direct_output.RegisterPageCallback(self.device_handle, self.onPage_closure)
        if result != S_OK:
            logging.warning("RegisterPageCallback failed")
            self.finish()
            raise DirectOutputError(result)


    def __del__(self, *args, **kwargs):
        logging.debug("DirectOutputDevice.__del__")
        self.finish()


    def finish(self):
        """
        De-initializes DLL. Must be called before program exit

        """
        if self.direct_output:
            logging.info("DirectOutputDevice deinitializing")
            self.direct_output.Deinitialize()
            self.direct_output = None
        else:
            logging.debug("nothing to do in finish()")


    def _OnDeviceClosure(self):
        """
        Returns a pointer to function that calls self._OnDevice method. This allows class methods to be called from within DirectOutput.dll

        http://stackoverflow.com/questions/7259794/how-can-i-get-methods-to-work-as-callbacks-with-python-ctypes

        """
        OnDevice_Proto = ctypes.WINFUNCTYPE(None, ctypes.c_void_p, ctypes.c_bool, ctypes.c_void_p)

        def func(hDevice, bAdded, pvContext):
            logging.info("device callback closure func: {}, {}, {}".format(hDevice, bAdded, pvContext))
            self._OnDevice(hDevice, bAdded, pvContext)

        return OnDevice_Proto(func)


    def _OnEnumerateClosure(self):
        """
        Returns a pointer to function that calls self._OnEnumerate method. This allows class methods to be called from within DirectOutput.dll

        http://stackoverflow.com/questions/7259794/how-can-i-get-methods-to-work-as-callbacks-with-python-ctypes

        """
        OnEnumerate_Proto = ctypes.WINFUNCTYPE(None, ctypes.c_void_p, ctypes.c_void_p)

        def func(hDevice, pvContext):
            logging.info("enumerate callback closure func: {}, {}".format(hDevice, pvContext))
            self._OnEnumerate(hDevice, pvContext)

        return OnEnumerate_Proto(func)


    def _OnPageClosure(self):
        """
        Returns a pointer to function that calls self._OnPage method. This allows class methods to be called from within DirectOutput.dll

        http://stackoverflow.com/questions/7259794/how-can-i-get-methods-to-work-as-callbacks-with-python-ctypes

        """
        OnPage_Proto = ctypes.WINFUNCTYPE(None, ctypes.c_void_p, ctypes.wintypes.DWORD, ctypes.c_bool, ctypes.c_void_p)

        def func(hDevice, dwPage, bActivated, pvContext):
            logging.info("page callback closure: {}, {}, {}, {}".format(hDevice, dwPage, bActivated, pvContext))
            self._OnPage(hDevice, dwPage, bActivated, pvContext)

        return OnPage_Proto(func)


    def _OnSoftButtonClosure(self):
        """
        Returns a pointer to function that calls self._OnSoftButton method. This allows class methods to be called from within DirectOutput.dll

        http://stackoverflow.com/questions/7259794/how-can-i-get-methods-to-work-as-callbacks-with-python-ctypes

        """
        OnSoftButton_Proto = ctypes.WINFUNCTYPE(None, ctypes.c_void_p, ctypes.wintypes.DWORD, ctypes.c_void_p)

        def func(hDevice, dwButtons, pvContext):
            logging.info("soft button callback closure: {}, {}, {}".format(hDevice, dwButtons, pvContext))
            self._OnSoftButton(hDevice, dwButtons, pvContext)

        return OnSoftButton_Proto(func)


    def _OnDevice(self, hDevice, bAdded, pvContext):
        """
        Internal function to register device handle

        """
        if not bAdded:
            raise NotImplementedError("Received a message that a device went away.")
        if self.device_handle and self.device_handle != hDevice:
            raise IndexError("Too many Saitek devices present")
        logging.info("_OnDevice")
        self.device_handle = hDevice


    def _OnEnumerate(self, hDevice, pvContext):
        """
        Internal function to process a device enumeration

        """
        logging.info("_OnEnumerate")
        self._OnDevice(hDevice, True, pvContext)


    def _OnPage(self, hDevice, dwPage, bActivated, pvContext):
        """
        Method called when page changes. Calls self.OnPage to hide hDevice and pvContext from end-user

        """
        logging.info("_OnPage")
        self.OnPage(dwPage, bActivated)


    def _OnSoftButton(self, hDevice, dwButtons, pvContext):
        """
        Method called when soft button changes. Calls self.OnSoftButton to hide hDevice and pvContext from end-user. Also hides change of page softbutton and press-up.

        """
        logging.info("_OnSoftButton")
        self.OnSoftButton(self.Buttons(dwButtons))


    def OnPage(self, page, activated):
        """
        Method called when a page changes. This should be overwritten by inheriting class

        Required Arguments:
        page -- page_id passed to AddPage
        activated -- true if this page has become the active page, false if this page was the active page

        """
        logging.info("OnPage({}, {})".format(page, activated))


    def OnSoftButton(self, buttons):
        """
        Method called when a soft button changes. This should be overwritten by inheriting class

        Required Arguments:
        buttons - Buttons object representing button state

        """
        logging.info("OnSoftButton({})".format(buttons))


    def SetProfile(self, profile):
        """
        Sets the profile used on the device.

        Required Arguments:
        device_handle -- ID of device
        profile -- full path of the profile to activate. passing None will clear the profile.
        """
        logging.debug("SetProfile({})".format(profile))
        return self.direct_output.SetProfile(self.device_handle, profile)


    def AddPage(self, page, name, active):
        """
        Adds a page to the MFD. If overriden by a derived class, you should
        call super().AddPage(*args, **kwargs)

        Required Arguments:
        page -- page ID to add
        name -- String specifying page name
        active -- True if page is to become the active page, if False this will not change the active page

        """
        logging.info("AddPage({}, {}, {})".format(page, name, active))
        self.direct_output.AddPage(self.device_handle, page, name, active)


    def RemovePage(self, page):
        """
        Removes a page from the MFD

        Required Arguments:
        page -- page ID to remove

        """
        logging.info("RemovePage({})".format(page))
        result = self.direct_output.RemovePage(self.device_handle, page)
        if result != S_OK:
            logging.error("RemovePage failed: {}".format(result))
            self.finish()
            raise DirectOutputError(result)


    def SetString(self, page, line, string):
        """
        Sets a string to display on the MFD

        Required Arguments:
        page -- the ID of the page to add the string to
        line -- the line to display the string on (0 = top, 1 = middle, 2 = bottom)
        string -- the string to display
        """
        logging.debug("SetString({}, {}, {})".format(page, line, string))
        result = self.direct_output.SetString(self.device_handle, page, line, string)
        if result != S_OK:
            logging.warning("SetString failed: {}".format(result))
            self.finish()
            raise DirectOutputError(result)


    def SetLed(self, page, led, value):
        """
        Sets LED state on a given page

        Required Arguments:
        page -- page number
        led -- ID of LED to change
        value -- value to set LED (1 = on, 0 = off)

        """
        logging.debug("SetLed({}, {}, {})".format(page, led, value))
        result = self.direct_output.SetLed(self.device_handle, page, led, value)
        if result != S_OK:
            logging.warning("SetLed failed: {}".format(result))
            self.finish()
            raise DirectOutputError(result)


class DeviceNotFoundError(Exception):

    def __str__(self):
        return "No Compatible Device Found"


class DLLError(Exception):

    def __init__(self, error_code):
        self.error_code = error_code
        if error_code == 126:
            self.msg = "specified file does not exist"
        elif error_code == 193:
            self.msg = "possible 32/64 bit mismatch between Python interpreter and DLL. Make sure you have installed both the 32- and 64-bit driver from Saitek's website"
        else:
            self.msg = "unspecified error"

    def __str__(self):
        return "Unable to load DirectOutput.dll - "+self.msg


class DirectOutputError(Exception):
    Errors = {
        E_HANDLE : "Invalid device handle specified.",
        E_INVALIDARG : "An argument is invalid, and I don't mean it has a poorly leg.",
        E_OUTOFMEMORY : "Download more RAM.",
        E_PAGENOTACTIVE : "Page not active, stupid page.",
        E_BUFFERTOOSMALL : "Buffer used was too small. Use a bigger buffer. See also E_OUTOFMEMORY.",
        E_NOTIMPL : "Feature not implemented, allegedly"
    }

    def __init__(self, error_code):
        self.error_code = error_code
        if error_code in self.Errors:
            self.msg = self.Errors[error_code]
        else:
            self.msg = "Unspecified DirectOutput Error - "+str(hex(error_code))

    def __str__(self):
        return self.msg


if __name__ == '__main__':
    # If you want it to go to a file?
    # logging.basicConfig(filename='directoutput.log', filemode='w', level=logging.DEBUG, format='%(asctime)s %(name)s [%(filename)s:%(lineno)d] %(message)s')
    # If you want less verbose logging?
    # logging.basicConfig(level=logging.INFO, format='%(asctime)s %(name)s [%(filename)s:%(lineno)d] %(message)s')
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(name)s [%(filename)s:%(lineno)d] %(message)s')

    import time, sys

    device = DirectOutputDevice(debug_level=1)
    print("Device initialized")

    device.AddPage(0, "Test", True)
    print("Test Page added")

    device.SetString(0, 0, "Test String")
    print("Test String added")

    device.AddPage(1, "Other", False)
    device.AddPage(2, "Another", False)

    while True:
        try:
            time.sleep(1)
        except:
            #This is used to catch Ctrl+C, calling finish method is *very* important to de-initalize device.
            device.finish()
            sys.exit()



from __future__ import absolute_import, with_statement, print_function, division, unicode_literals

import os
import traceback
import sys
_ROOT = os.path.abspath(os.path.dirname(__file__))

class TradeEnv(object):
    """
        Container for a TradeDangerous "environment", which is a
        collection of operational parameters.

        To print debug lines, use DEBUG<N>, e.g. DEBUG0, which
        takes a format() string and parameters, e.g.

            DEBUG1("hello, {world}{}", "!", world="world")

        is equivalent to:

            if tdenv.debug >= 1:
                print("#hello, {world}{}".format(
                        "!", world="world"
                ))

        Use "NOTE" to print remarks which can be disabled with -q.
    """


    defaults = {
        'debug': 0,
        'detail': 0,
        'quiet': 0,
        'dataDir': os.environ.get('TD_DATA') or os.path.join(os.getcwd(), 'data'),
        'tmpDir': os.environ.get('TD_TMP') or os.path.join(os.getcwd(), 'tmp'),
        'templateDir': os.environ.get('TD_TEMPLATES', None) or os.path.join(_ROOT, 'templates'),
        'cwDir': os.getcwd()
    }


    encoding = sys.stdout.encoding
    if str(sys.stdout.encoding).upper() != 'UTF-8':
        def uprint(self, *args, **kwargs):
            try:
                print(*args, **kwargs)
            except UnicodeEncodeError as e:
                if not self.quiet:
                    print(
                        "CAUTION: Your terminal/console couldn't handle some "
                        "text I tried to print."
                    )
                    if 'EXCEPTIONS' in os.environ:
                        traceback.print_exc()
                    else:
                        print(str(e))
                strs = [
                    str(arg).
                        encode(TradeEnv.encoding, errors='replace').
                        decode(TradeEnv.encoding)
                    for arg in args
                ]
                print(*strs, **kwargs)
    else:
        uprint = print


    def __init__(self, properties=None, **kwargs):
        properties = properties or dict()
        self.__dict__.update(self.defaults)
        if properties:
            self.__dict__.update(properties.__dict__)
        if kwargs:
            self.__dict__.update(kwargs)


    def __getattr__(self, key):
        """ Return the default for attributes we don't have """
        if key.startswith("DEBUG"):
            # Self-assembling DEBUGN functions
            def __DEBUG_ENABLED(outText, *args, **kwargs):
                print('#', outText.format(*args, **kwargs))

            def __DEBUG_DISABLED(*args, **kwargs):
                pass

            # Tried to call a .DEBUG<N> function which hasn't
            # been called before; create a stub.
            debugLevel = int(key[5:])
            if self.debug > debugLevel:
                debugFn = __DEBUG_ENABLED
            else:
                debugFn = __DEBUG_DISABLED
            setattr(self, key, debugFn)
            return debugFn

        if key == "NOTE":
            def __NOTE_ENABLED(outText, *args, file=None, **kwargs):
                self.uprint(
                    "NOTE:", str(outText).format(*args, **kwargs),
                    file=file,
                )

            def __NOTE_DISABLED(*args, **kwargs):
                pass

            # Tried to call "NOTE" but it hasn't been called yet,
            if not self.quiet:
                noteFn = __NOTE_ENABLED
            else:
                noteFn = __NOTE_DISABLED
            setattr(self, key, noteFn)
            return noteFn

        if key == "WARN":
            def _WARN_ENABLED(outText, *args, file=None, **kwargs):
                self.uprint(
                    "WARNING:", str(outText).format(*args, **kwargs),
                    file=file
                )

            def _WARN_DISABLED(*args, **kwargs):
                pass

            noteFn = _WARN_DISABLED if self.quiet > 1 else _WARN_ENABLED
            setattr(self, key, noteFn)
            return noteFn

        return None


from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from .exceptions import *
from .parsing import *
from ..tradedb import TradeDB
from ..tradecalc import TradeCalc, Route

import math

######################################################################
# Parser config

help='Find places to buy a given item within range of a given station.'
name='trade'
epilog=None
wantsTradeDB=True
arguments = [
    ParseArgument(
        'origin',
        help='Station you are purchasing from.',
        type=str,
    ),
    ParseArgument(
        'dest',
        help='Station you are selling to.',
        type=str,
    ),
]
switches = [
]

######################################################################
# Perform query and populate result set

def run(results, cmdenv, tdb):
    from .commandenv import ResultRow

    calc = TradeCalc(tdb, cmdenv)

    lhs = cmdenv.startStation
    rhs = cmdenv.stopStation

    if lhs == rhs:
        raise CommandLineError("Must specify two different stations.")

    results.summary = ResultRow()
    results.summary.fromStation = lhs
    results.summary.toStation = rhs

    trades = calc.getTrades(lhs, rhs)
    if not trades:
        raise CommandLineError("No profitable trades {} -> {}".format(
            lhs.name(), rhs.name()
        ))

    if cmdenv.detail > 1:
        tdb.getAverageSelling()
        tdb.getAverageBuying()

    for item in trades:
        results.rows.append(item)

    return results

#######################################################################
## Transform result set into output

def render(results, cmdenv, tdb):
    from ..formatting import RowFormat, ColumnFormat, max_len

    longestNameLen = max_len(results.rows, key=lambda row: row.item.name(cmdenv.detail))

    rowFmt = RowFormat()
    rowFmt.addColumn('Item', '<', longestNameLen,
            key=lambda row: row.item.name(cmdenv.detail))
    rowFmt.addColumn('Profit', '>', 10, 'n',
            key=lambda row: row.gainCr)
    rowFmt.addColumn('Cost', '>', 10, 'n',
            key=lambda row: row.costCr)
    if cmdenv.detail > 1:
        rowFmt.addColumn('AvgCost', '>', 10,
            key=lambda row: tdb.avgSelling.get(row.item.ID, 0)
        )
        rowFmt.addColumn('Buying', '>', 10,
            key=lambda row: row.costCr + row.gainCr
        )
        rowFmt.addColumn('AvgBuy', '>', 10,
            key=lambda row: tdb.avgBuying.get(row.item.ID, 0)
        )

    if cmdenv.detail:
        rowFmt.addColumn('Supply', '>', 10,
            key=lambda row: '{:n}'.format(row.supply) if row.supply >= 0 else '?')
        rowFmt.addColumn('Demand', '>', 10,
            key=lambda row: '{:n}'.format(row.demand) if row.demand >= 0 else '?')
        rowFmt.addColumn('SrcAge', '>', 8, '.2f',
            key=lambda row: (row.srcAge / 86400))
        rowFmt.addColumn('DstAge', '>', 8, '.2f',
            key=lambda row: (row.dstAge / 86400))

    if not cmdenv.quiet:
        heading, underline = rowFmt.heading()
        print(heading, underline, sep='\n')

    for row in results.rows:
        print(rowFmt.format(row))